'''version 1.1'''

import random, sys, pickle, collections, operator, itertools, time, math, os
import importlib
from .util import *
from collections import defaultdict
import numpy as np
import math
from copy import copy
import argparse
import os
from dataclasses import dataclass
from typing import Callable
import concurrent.futures
from memory_profiler import profile
from time import sleep
from copy import copy, deepcopy

# Optional dependency
if importlib.util.find_spec("progress") is not None:
    from progress.bar import Bar
else:
    Bar = None

def _promocalc_progress_callback(tasks: int) -> Callable:
    # Update the progress bar
    if Bar is not None:
        bar = Bar('Running Promoter Predictions: ', max=tasks)
    else:
        return None
    max = tasks
    itterations = 0

    def _callback():
        nonlocal itterations
        itterations += 1
        nonlocal max
        if itterations == max:
            bar.next()
            bar.finish()
            return None

        bar.next()
        return None


    return _callback


# k and BETA for La Fleur dataset
LOGK   = -2.80271176
BETA    = 0.81632623

@dataclass
class PromoCalcResults:
    """Class to hold results of the promoter prediction"""
    promoter_sequence: str
    TSS: int
    UP: str
    hex35: str
    spacer: str
    hex10: str
    disc: str
    ITR: str
    dG_total: float
    dG_10: float
    dG_35: float
    dG_disc: float
    dG_ITR: float
    dG_ext10: float
    dG_spacer: float
    dG_bind: float
    dG_UP: float
    Tx_rate: float
    UP_position: int
    hex35_position: int
    spacer_position: int
    hex10_position: int
    disc_position: int
    fraction: int
    strand: str
    drop: bool
    length: int

    def __getitem__(self, key):
        if key in self.__annotations__:
            return getattr(self, key)
        else:
            raise KeyError(key)

    def __setitem__(self, key, value):
        setattr(self, key, value)

    def keys(self):
        return self.__annotations__.keys()

def unpickler(infile):
    """ Unpickle a file """
    with open(infile, 'rb') as handle:
        obj= pickle.load(handle)
    return obj

def _revcomp(seq):
    """ Reverse complement a sequence """
    revcomp = {'U' : 'A', 'A' : 'T', 'G' : 'C', 'T' : 'A', 'C' : 'G'}
    return "".join([revcomp[letter] for letter in seq[::-1] ])

def get_matrices(two_mer_encoder, three_mer_encoder, spacer_encoder, coeffs):
    """ Get matrices for linear model """

    #Extract dG values from model coefficients
    ref10_0 = coeffs.tolist()[0:64]
    ref10_3 = coeffs.tolist()[64:128]
    ref35_0 = coeffs.tolist()[128:192]
    ref35_3 = coeffs.tolist()[192:256]
    discs   = coeffs.tolist()[256:256+64]
    x10     = coeffs.tolist()[256+64:256+64+16]
    spacs   = coeffs.tolist()[256+64+16:256+64+16+3]

    # make dG matrices for each feature
    dg10_0  = get_dg_matrices(ref10_0,three_mer_encoder)
    dg10_3  = get_dg_matrices(ref10_3,three_mer_encoder)
    dg35_0  = get_dg_matrices(ref35_0,three_mer_encoder)
    dg35_3  = get_dg_matrices(ref35_3,three_mer_encoder)
    dmers   = get_dg_matrices(discs,three_mer_encoder)
    x10mers = get_dg_matrices(x10,two_mer_encoder)
    spacers = get_dg_matrices(spacs, spacer_encoder)

    return dg10_0, dg10_3, dg35_0, dg35_3, dmers, x10mers, spacers

def scan_arbitrary(inputSequence, model, inters, constraints, dg10_0, dg10_3, dg35_0, dg35_3, dmers, x10mers, spacers):
    """Scan sequence left to right with no TSS information. Calc dG of all possible promoter configurations."""
    seq_query = {}
    upstream = constraints[0]
    downstream = constraints[1]
    sequence = upstream + inputSequence + downstream

    # first 20 nt will be initial UP candidate
    for i in range(0,len(sequence)):
        tempUP = sequence[i:i+24]
        temp35 = sequence[i+25:25+i+6]  # leaves 1 nt between h35 and UPs

        # bounds defined by what was present during parameterization
        for j in range(15,21):
            tempspacer = sequence[i+25+6:25+i+6+j]
            temp10     = sequence[25+i+j+6:25+i+j+12]
            for k in range(6,11):
                tempdisc  = sequence[25+i+j+12:25+i+j+12+k]
                tempITR   =sequence[25+i+j+12+k:45+i+j+12+k]
                if len(tempITR) < 20:
                    continue
                else:
                    model_results= linear_free_energy_model(tempUP,
                                                            temp35,
                                                            tempspacer,
                                                            temp10,
                                                            tempdisc,
                                                            tempITR,
                                                            dg10_0,
                                                            dg10_3,
                                                            dg35_0,
                                                            dg35_3,
                                                            dmers,
                                                            x10mers,
                                                            spacers,
                                                            model,
                                                            inters)
                    TSS_distance = i + len(tempUP) + len(temp35) + len(tempspacer) + len(temp10) + len(tempdisc)
                    # seq_query[(float(dG_bind), float(dG_total), TSS_distance)] = ((tempUP, temp35, tempspacer, temp10, tempdisc, tempITR),(dg10, dg35, dg_disc, dg_ITR, dg_ext10, dg_spacer, dg_UP))
                    seq_query[(float(model_results.dg_total), float(model_results.dg_apparent), TSS_distance)] = ((tempUP, temp35, tempspacer, temp10, tempdisc, tempITR),(model_results.dg_10, model_results.dg_35, model_results.dg_disc, model_results.dg_ITR, model_results.dg_ext10, model_results.dg_spacer, model_results.dg_UP))

    print("best: ", min(seq_query.items(), key=operator.itemgetter(0)))

    best = (collections.OrderedDict(sorted(seq_query.items())), min(seq_query.items(), key=operator.itemgetter(0)))
    return best, seq_query

def linear_free_energy_model(UP, h35, spacer, h10, disc, ITR, dg10_0, dg10_3, dg35_0, dg35_3, dmers, x10mers, spacers, coeffs, inters):
    """ Calculate dG of a promoter configuration """

    prox_UP = UP[-int(len(UP)/2)::]
    dist_UP = UP[0:int(len(UP)/2)]

    # CATEGORICAL FEATURES
    ext10           = spacer[-3:-1] # TGN motif, contacts sigma
    hex10_0         = h10[0:3]
    hex10_3         = h10[3::]
    hex35_0         = h35[0:3]
    hex35_3         = h35[3::]
    disc_first_3mer = disc[0:3]
    spacer_length   = str(len(spacer))

    # NUMERICAL FEATURES
    _, _, dg_ITR     = calc_DNA_RNA_hybrid_energy(ITR) # calc R-loop strength
    rigidity                 = calc_rigidity(seq = UP + h35 + spacer[0:14])

    width_proxy_prox = calc_groove_width(prox_UP)
    width_proxy_dist = calc_groove_width(dist_UP)

    # NORMALIZE NUMERICAL FEATURES BY MAX IN TRAINING SET
    numericals         = np.array([width_proxy_dist, width_proxy_prox, dg_ITR, rigidity])
    normalizing_values = [256.0, 255.0, 4.300000000000002, 25.780434782608694]
    numerical_coefs    = np.array(coeffs.tolist()[-4::])
    normald            = np.divide(numericals,normalizing_values)
    dg_numerical       = np.multiply(normald, numerical_coefs)
    dg10      = dg10_0[hex10_0] + dg10_3[hex10_3]
    dg35      = dg35_0[hex35_0] + dg35_3[hex35_3]
    dg_disc   = dmers[disc_first_3mer]
    dg_ITR    = dg_numerical[-2]
    dg_ext10  = x10mers[ext10]

    x = float(spacer_length)
    dg_spacer = 0.1463*x**2 - 4.9113*x + 41.119

    dg_UP        = dg_numerical[0] + dg_numerical[1] + dg_numerical[-1]
    dG_apparent  = (dg10 + dg35 + dg_disc + dg_ITR + dg_ext10 + dg_spacer + dg_UP + inters[0] - LOGK)/BETA
    dG_total     = dg10 + dg35 + dg_disc + dg_ITR + dg_ext10 + dg_spacer + dg_UP + inters[0]

    @dataclass
    class PromoModelResults:
        """Class to hold results of the linear free energy model"""
        dg_total: float
        dg_apparent: float
        dg_10: float
        dg_35: float
        dg_disc: float
        dg_ITR: float
        dg_ext10: float
        dg_spacer: float
        dg_UP: float

    return PromoModelResults(dG_total,  # @TODO: Univy this with PromoPredictionResults
                             dG_apparent,
                             dg10,
                             dg35,
                             dg_disc,
                             dg_ITR,
                             dg_ext10,
                             dg_spacer,
                             dg_UP)

def predict(sequence, constraints):
    """ Predict the free energy of a promoter configuration """

    # Initialize model and matrices
    install_location = os.path.dirname(os.path.realpath(__file__))
    layer1 = np.load(install_location + '/free_energy_coeffs.npy')
    inters = np.load(install_location + '/model_intercept.npy')

    two_mer_encoder   = kmer_encoders(k = 2)
    three_mer_encoder = kmer_encoders(k = 3)
    spacer_encoder    = length_encoders(16, 18)
    dg10_0, dg10_3, dg35_0, dg35_3, dmers, x10mers, spacers = get_matrices(two_mer_encoder = two_mer_encoder, three_mer_encoder = three_mer_encoder, spacer_encoder = spacer_encoder, coeffs = layer1)

    # Scan DNA and return predictions
    (_, result), query   = scan_arbitrary(inputSequence = sequence,
                                            model = layer1, inters = inters, constraints = constraints, dg10_0 = dg10_0, dg10_3 = dg10_3,
                                            dg35_0 =dg35_0, dg35_3 = dg35_3, dmers = dmers , x10mers = x10mers, spacers = spacers)

    @dataclass
    class PromoPredictResults:
        """Class to hold results of the promoter prediction"""
        dg_total: float
        query: dict
        UP: str
        h35: str
        spacer: str
        h10: str
        disc: str
        ITR: str

    dG_total, UP, h35, spacer, h10, disc, ITR = result[0][0], result[1][0][0],result[1][0][1],result[1][0][2],result[1][0][3], result[1][0][4], result[1][0][5]
    return PromoPredictResults(dG_total, query, UP, h35, spacer, h10, disc, ITR)

class Promoter_Calculator(object):
    """ Class to calculate the free energy of a promoter configuration """

    def __init__(self, organism = 'Escherichia coli str. K-12 substr. MG1655',
                       sigmaLevels = {'70' : 1.0, '19' : 0.0, '24' : 0.0, '28' : 0.0, '32' : 0.0, '38' : 0.0, '54' : 0.0},
                       threads=1,
                       verbosity=1):
        # Initialize model and matrices
        self.threads = threads
        self.verbosity = verbosity
        path = os.path.dirname(os.path.abspath(__file__))
        self.layer1 = np.load(path + '/free_energy_coeffs.npy')
        self.inters = np.load(path + '/model_intercept.npy')
        self.two_mer_encoder   = kmer_encoders(k = 2)
        self.three_mer_encoder = kmer_encoders(k = 3)
        self.spacer_encoder    = length_encoders(16, 18)
        self.dg10_0, self.dg10_3, self.dg35_0, self.dg35_3, self.dmers, self.x10mers, self.spacers = get_matrices(two_mer_encoder = self.two_mer_encoder, three_mer_encoder = self.three_mer_encoder, spacer_encoder = self.spacer_encoder, coeffs = self.layer1)
        self.model = self.layer1
        self.organism = organism
        self.sigmaLevels = sigmaLevels

        if organism == 'in vitro':
            self.K = 42.00000
            self.BETA = 0.81632623
        elif organism == 'Escherichia coli str. K-12 substr. MG1655':
            self.K = 42.00000
            self.BETA = 1.636217004872062
        else:
            self.K = 42.00000
            self.BETA = 1.636217004872062

    # Identify promoter with minimum dG_total (across many possible promoter states) for each TSS position in an inputted sequence.
    def predict(self, sequence, TSS_range, callback=None, min_only=True):
        """ Predict the free energy of a promoter configuration """

        min_states, all_states = self.parallelizer(sequence, TSS_range,
                                                   callback=callback,
                                                   verbosity=self.verbosity,
                                                   min_only=min_only)

        return (min_states, all_states)

 
    def parallelizer(self, sequence, TSS_range, callback=None, min_only=True, verbosity=1):
        threads = self.threads

        Min_States = {}
        All_States = {}

        target_tss = range(TSS_range[0], TSS_range[1])

        model_params = {'dg10_0' : self.dg10_0,
                        'dg10_3' : self.dg10_3,
                        'dg35_0' : self.dg35_0,
                        'dg35_3' : self.dg35_3,
                        'dmers' : self.dmers,
                        'x10mers' : self.x10mers,
                        'spacers' : self.spacers,
                        'model' : self.model,
                        'inters' : self.inters,
                        'K' : self.K,
                        'BETA' : self.BETA,}

        def _next_position():
            sample_region = 200
            for next_TSS in target_tss:
                if len(sequence) < sample_region:
                    locale = next_TSS
                    region = sequence
                elif next_TSS < sample_region/2:
                    locale = next_TSS
                    region = sequence[:sample_region]
                elif next_TSS > len(sequence) - sample_region/2:
                    locale = next_TSS - (len(sequence) - sample_region)
                    region = sequence[len(sequence)-sample_region:]
                else:
                    locale = int(sample_region/2)
                    region = sequence[int(next_TSS-(sample_region/2)) : int(next_TSS+(sample_region/2))]
                yield (region, next_TSS, model_params, locale)

        sleep_time = 0.01
        sleep_interval = 0.0001
        if threads > 1:
            with concurrent.futures.ProcessPoolExecutor(max_workers=threads) as multiprocessor:
                futures: list = []
                for task in _next_position():
                    future_object = multiprocessor.submit(self.worker, *task)
                    future_object.add_done_callback(lambda x: callback())
                    futures.append(future_object)
                    while len(multiprocessor._pending_work_items) > 5*threads and sleep_time:
                        sleep(sleep_time)
                        if len(multiprocessor._pending_work_items) == 0:
                            sleep_time -= sleep_interval
                        continue
                parallel_output = [future.result() for future in futures]
        else:
            parallel_output: list = []
            for task in _next_position():
                result = self.worker(*task)
                parallel_output.append(result)
                callback()

        # Unpack results

        parallel_output = [output for output in parallel_output if output]  # Remove empty results
        
        for result_group in parallel_output:
            for hit in result_group:
                TSS, DISC_length, SPACER_length, result = hit
                All_States[TSS] = {}
                All_States[TSS][ (DISC_length, SPACER_length) ] = result
                if TSS in Min_States:
                    if result['dG_total'] < Min_States[TSS]['dG_total']:  Min_States[TSS] = result
                else:
                    Min_States[TSS] = result

        return (Min_States, All_States)

    @staticmethod
    def worker(sequence, TSS, model_params, locale, min_only=True):
        results = []

        UPS_length = 24
        HEX35_length = 6
        UPS_HEX35_SPACER = 1
        SPACER_length_range = [15, 21]
        HEX10_length = 6
        DISC_length_range = [6, 11]
        ITR_length = 20
        All_States = {}
        Min_States = {}

        

        for DISC_length in range(DISC_length_range[0],DISC_length_range[1]):
            if locale - DISC_length >= 0 and locale + ITR_length <= len(sequence):
                tempdisc = sequence[ locale - DISC_length : locale  ]
                tempITR  = sequence[ locale : locale + ITR_length]

                for SPACER_length in range(SPACER_length_range[0], SPACER_length_range[1]):

                    if locale - DISC_length - HEX10_length - SPACER_length - HEX35_length - UPS_length - UPS_HEX35_SPACER >= 0:
                        temp10     = sequence[ locale - DISC_length - HEX10_length : locale - DISC_length]
                        tempspacer = sequence[ locale - DISC_length - HEX10_length - SPACER_length : locale - DISC_length - HEX10_length ]
                        temp35     = sequence[ locale - DISC_length - HEX10_length - SPACER_length - HEX35_length : locale - DISC_length - HEX10_length - SPACER_length]
                        tempUP     = sequence[ locale - DISC_length - HEX10_length - SPACER_length - HEX35_length - UPS_length - UPS_HEX35_SPACER:  locale - DISC_length - HEX10_length - SPACER_length - HEX35_length - UPS_HEX35_SPACER]

                        model_results = linear_free_energy_model(tempUP,
                                                                    temp35,
                                                                    tempspacer,
                                                                    temp10,
                                                                    tempdisc,
                                                                    tempITR,
                                                                    model_params['dg10_0'],
                                                                    model_params['dg10_3'],
                                                                    model_params['dg35_0'],
                                                                    model_params['dg35_3'],
                                                                    model_params['dmers'],
                                                                    model_params['x10mers'],
                                                                    model_params['spacers'],
                                                                    model_params['model'],
                                                                    model_params['inters'])
                        dG_bind = (model_results.dg_10 +
                                    model_results.dg_35 +
                                    model_results.dg_spacer +
                                    model_results.dg_ext10 +
                                    model_results.dg_UP)

                        Tx_rate = float(model_params['K'] * math.exp(- model_params['BETA'] * model_results.dg_total ))
                        
                        result = {'promoter_sequence' : sequence[locale - DISC_length - HEX10_length - SPACER_length - HEX35_length - UPS_length - UPS_HEX35_SPACER : locale + ITR_length ],
                                    'TSS' : TSS,
                                    'UP' : tempUP, 
                                    'hex35' : temp35, 
                                    'spacer' : tempspacer, 
                                    'hex10' : temp10, 
                                    'disc' : tempdisc, 
                                    'ITR' : tempITR,
                                    'dG_total' : copy(model_results.dg_total), 
                                    'dG_10' : copy(model_results.dg_10), 
                                    'dG_35' : copy(model_results.dg_35), 
                                    'dG_disc' : copy(model_results.dg_disc), 
                                    'dG_ITR' : copy(model_results.dg_ITR), 
                                    'dG_ext10' : copy(model_results.dg_ext10), 
                                    'dG_spacer' : copy(model_results.dg_spacer), 
                                    'dG_UP' : copy(model_results.dg_UP), 
                                    'dG_bind' : copy(dG_bind),
                                    'Tx_rate' : copy(Tx_rate),
                                    'UP_position' : [TSS - DISC_length - HEX10_length - SPACER_length - HEX35_length - UPS_HEX35_SPACER - UPS_length,
                                                    TSS - DISC_length - HEX10_length - SPACER_length - HEX35_length - UPS_HEX35_SPACER],
                                    'hex35_position' : [TSS - DISC_length - HEX10_length - SPACER_length - HEX35_length, TSS - DISC_length - HEX10_length - SPACER_length],
                                    'spacer_position' : [TSS - DISC_length - HEX10_length - SPACER_length, TSS - DISC_length - HEX10_length],
                                    'hex10_position' : [TSS - DISC_length - HEX10_length, TSS - DISC_length],
                                    'disc_position' : [TSS - DISC_length, TSS]
                                    }
                        


                        
                        result = PromoCalcResults(**result, fraction=None, strand=None, drop=False, length = False)
                        results.append([TSS, DISC_length, SPACER_length, result])

        if not results:
            return results
        
        if min_only:
            min_result = None
            for TSS, DISC_length, SPACER_length, finding in results:
                if min_result:
                    if finding['dG_total'] < min_result[3]['dG_total']:
                        min_result = [TSS, DISC_length, SPACER_length, finding]
                else:
                    min_result = [TSS, DISC_length, SPACER_length, finding]
            return [min_result]
        
        return results

    def run(self, sequence, TSS_range = None, callback=None):
        """ Run the predictor """

        if TSS_range is None: TSS_range = [0, len(sequence)]

        self.sequence = sequence
        self.TSS_range = TSS_range
        self.TSS_range_rev = [len(sequence) - TSS_range[1], len(sequence) - TSS_range[0]]

        # print "self.TSS_range_rev: ", self.TSS_range_rev

        if not callback:
            callback = lambda: None

        fwd_sequence = sequence
        rev_sequence = _revcomp(sequence)
        (Forward_Min_States, Forward_All_States) = self.predict(fwd_sequence, TSS_range = self.TSS_range, callback=callback)
        (Reverse_Min_States_Temp, Reverse_All_States_Temp) = self.predict(rev_sequence, TSS_range = self.TSS_range_rev, callback=callback)

        Reverse_Min_States = {}
        Reverse_All_States = {}
        # 0  <------>500 fwd 500 bp
        # 500<------>0   rev 500 bp
        #      275 TSS
        #      200-300
        #500-275 = 225
        #

        for TSS in Reverse_Min_States_Temp.keys():
            Reverse_Min_States[len(sequence) - TSS] = Reverse_Min_States_Temp[TSS]
            #Reverse_All_States[len(sequence) - TSS] = Reverse_All_States_Temp[TSS]

        self.Forward_Predictions_per_TSS = Forward_Min_States
        self.Reverse_Predictions_per_TSS = Reverse_Min_States

    def output(self):
        """ Output the results as a dictionary """
        output = {'organism' : self.organism,
                  'sigmaLevels' : self.sigmaLevels,
                  'K' : self.K,
                  'beta' : self.BETA,
                  'sequence' : self.sequence,
                  'TSS_range' : self.TSS_range,
                  'Forward_Predictions_per_TSS' : self.Forward_Predictions_per_TSS,
                  'Reverse_Predictions_per_TSS' : self.Reverse_Predictions_per_TSS
                }
        return output.copy()




import util, random, sys, pickle, collections, operator, itertools, time, math, os
from collections import defaultdict
import numpy as np
from rich import print as rprint

# k and BETA for La Fleur dataset
LOGK   = -2.80271176
BETA    = 0.81632623

def unpickler(infile):
    with open(infile, 'rb') as handle:
        obj= pickle.load(handle)
    return obj

def _revcomp(seq):
    revcomp = {'U' : 'A', 'A' : 'T', 'G' : 'C', 'T' : 'A', 'C' : 'G'}
    return "".join([revcomp[letter] for letter in seq[::-1] ])

def get_matrices(two_mer_encoder, three_mer_encoder, spacer_encoder, coeffs):

    #Extract dG values from model coefficients
    ref10_0 = coeffs.tolist()[0:64]
    ref10_3 = coeffs.tolist()[64:128]
    ref35_0 = coeffs.tolist()[128:192]
    ref35_3 = coeffs.tolist()[192:256]
    discs   = coeffs.tolist()[256:256+64]
    x10     = coeffs.tolist()[256+64:256+64+16]
    spacs   = coeffs.tolist()[256+64+16:256+64+16+3]

    # make dG matrices for each feature
    dg10_0  = util.get_dg_matrices(ref10_0,three_mer_encoder)
    dg10_3  = util.get_dg_matrices(ref10_3,three_mer_encoder)
    dg35_0  = util.get_dg_matrices(ref35_0,three_mer_encoder)
    dg35_3  = util.get_dg_matrices(ref35_3,three_mer_encoder)
    dmers   = util.get_dg_matrices(discs,three_mer_encoder)
    x10mers = util.get_dg_matrices(x10,two_mer_encoder)
    spacers = util.get_dg_matrices(spacs, spacer_encoder)

    return dg10_0, dg10_3, dg35_0, dg35_3, dmers, x10mers, spacers

# Scan sequence left to right with no TSS information. Calc dG of all possible promoter configurations.
def scan_arbitrary(inputSequence, two_mer_encoder, three_mer_encoder, model, inters, constraints, dg10_0, dg10_3, dg35_0, dg35_3, dmers, x10mers, spacers):
    seq_query = {}
    upstream = constraints[0]
    downstream = constraints[1]
    sequence = upstream + inputSequence + downstream

    # first 20 nt will be initial UP candidate
    for i in range(0,len(sequence)):
        tempUP = sequence[i:i+24]
        temp35 = sequence[i+25:25+i+6]  # leaves 1 nt between h35 and UPs

        # bounds defined by what was present during parameterization
        for j in range(15,21):
            tempspacer = sequence[i+25+6:25+i+6+j]
            temp10     = sequence[25+i+j+6:25+i+j+12]
            for k in range(6,11):
                tempdisc  = sequence[25+i+j+12:25+i+j+12+k]
                tempITR   =sequence[25+i+j+12+k:45+i+j+12+k]
                if len(tempITR) < 20:
                    continue
                else:
                    dG_total, dG_apparent, dg10, dg35, dg_disc, dg_ITR, dg_ext10, dg_spacer, dg_UP= linear_free_energy_model(tempUP, temp35, tempspacer, temp10, tempdisc, tempITR, dg10_0, dg10_3, dg35_0, dg35_3, dmers, x10mers, spacers, model, inters)
                    dG_bind  = dg10 + dg35 + dg_spacer + dg_ext10 + dg_UP
                    # dG_bind  = dg10 + dg_ext10 + dg_spacer + dg_UP
                    TSS_distance = i + len(tempUP) + len(temp35) + len(tempspacer) + len(temp10) + len(tempdisc)
                    # seq_query[(float(dG_bind), float(dG_total), TSS_distance)] = ((tempUP, temp35, tempspacer, temp10, tempdisc, tempITR),(dg10, dg35, dg_disc, dg_ITR, dg_ext10, dg_spacer, dg_UP))
                    seq_query[(float(dG_total), float(dG_apparent), TSS_distance)] = ((tempUP, temp35, tempspacer, temp10, tempdisc, tempITR),(dg10, dg35, dg_disc, dg_ITR, dg_ext10, dg_spacer, dg_UP))

    print("best: ", min(seq_query.items(), key=operator.itemgetter(0)))

    best = (collections.OrderedDict(sorted(seq_query.items())), min(seq_query.items(), key=operator.itemgetter(0)))
    return best, seq_query

def linear_free_energy_model(UP, h35, spacer, h10, disc, ITR, dg10_0, dg10_3, dg35_0, dg35_3, dmers, x10mers, spacers, coeffs, inters):

    prox_UP = UP[-int(len(UP)/2)::]
    dist_UP = UP[0:int(len(UP)/2)]

    # CATEGORICAL FEATURES
    ext10           = spacer[-3:-1] # TGN motif, contacts sigma
    hex10_0         = h10[0:3]
    hex10_3         = h10[3::]
    hex35_0         = h35[0:3]
    hex35_3         = h35[3::]
    disc_first_3mer = disc[0:3]
    spacer_length   = str(len(spacer))

    # NUMERICAL FEATURES
    dg_dna,dg_rna,dg_ITR     = util.calc_DNA_RNA_hybrid_energy(ITR) # calc R-loop strength
    rigidity                 = util.calc_rigidity(seq = UP + h35 + spacer[0:14])

    width_proxy_prox = util.calc_groove_width(prox_UP)
    width_proxy_dist = util.calc_groove_width(dist_UP)

    # NORMALIZE NUMERICAL FEATURES BY MAX IN TRAINING SET
    numericals         = np.array([width_proxy_dist, width_proxy_prox, dg_ITR, rigidity])
    normalizing_values = [256.0, 255.0, 4.300000000000002, 25.780434782608694]
    numerical_coefs    = np.array(coeffs.tolist()[-4::])
    normald            = np.divide(numericals,normalizing_values)
    dg_numerical       = np.multiply(normald, numerical_coefs)
    dg10      = dg10_0[hex10_0] + dg10_3[hex10_3]
    dg35      = dg35_0[hex35_0] + dg35_3[hex35_3]
    dg_disc   = dmers[disc_first_3mer]
    dg_ITR    = dg_numerical[-2]
    dg_ext10  = x10mers[ext10]

    x = float(spacer_length)
    dg_spacer = 0.1463*x**2 - 4.9113*x + 41.119

    dg_UP        = dg_numerical[0] + dg_numerical[1] + dg_numerical[-1]
    dG_apparent  = (dg10 + dg35 + dg_disc + dg_ITR + dg_ext10 + dg_spacer + dg_UP + inters[0] - LOGK)/BETA
    dG_total     = dg10 + dg35 + dg_disc + dg_ITR + dg_ext10 + dg_spacer + dg_UP + inters[0]

    return dG_total, dG_apparent, dg10, dg35, dg_disc, dg_ITR, dg_ext10, dg_spacer, dg_UP

def predict(sequence, constraints):

    # Initialize model and matrices
    layer1 = np.load('free_energy_coeffs.npy')
    inters = np.load('model_intercept.npy')

    two_mer_encoder   = util.kmer_encoders(k = 2)
    three_mer_encoder = util.kmer_encoders(k = 3)
    spacer_encoder    = util.length_encoders(16, 18)
    dg10_0, dg10_3, dg35_0, dg35_3, dmers, x10mers, spacers = get_matrices(two_mer_encoder = two_mer_encoder, three_mer_encoder = three_mer_encoder, spacer_encoder = spacer_encoder, coeffs = layer1)

    # Scan DNA and return predictions
    (od,result), query   = scan_arbitrary(inputSequence = sequence, two_mer_encoder = two_mer_encoder, three_mer_encoder = three_mer_encoder,
                                            model = layer1, inters = inters, constraints = constraints, dg10_0 = dg10_0, dg10_3 = dg10_3,
                                            dg35_0 =dg35_0, dg35_3 = dg35_3, dmers = dmers , x10mers = x10mers, spacers = spacers)

    dG_total, UP, h35, spacer, h10, disc, ITR = result[0][0], result[1][0][0],result[1][0][1],result[1][0][2],result[1][0][3],result[1][0][4],result[1][0][5]

    return dG_total, query, UP, h35, spacer, h10, disc, ITR

class Promoter_Calculator(object):

    def __init__(self, organism = 'Escherichia coli str. K-12 substr. MG1655',
                       sigmaLevels = {'70' : 1.0, '19' : 0.0, '24' : 0.0, '28' : 0.0, '32' : 0.0, '38' : 0.0, '54' : 0.0}):

        # Initialize model and matrices
        path = os.path.dirname(os.path.abspath(__file__))
        self.layer1 = np.load(path + '/free_energy_coeffs.npy')
        self.inters = np.load(path + '/model_intercept.npy')

        self.two_mer_encoder   = util.kmer_encoders(k = 2)
        self.three_mer_encoder = util.kmer_encoders(k = 3)
        self.spacer_encoder    = util.length_encoders(16, 18)
        self.dg10_0, self.dg10_3, self.dg35_0, self.dg35_3, self.dmers, self.x10mers, self.spacers = get_matrices(two_mer_encoder = self.two_mer_encoder, three_mer_encoder = self.three_mer_encoder, spacer_encoder = self.spacer_encoder, coeffs = self.layer1)

        self.model = self.layer1
        self.organism = organism
        self.sigmaLevels = sigmaLevels

        if organism == 'in vitro':
            self.K = 42.00000
            self.BETA = 0.81632623
        elif organism == 'Escherichia coli str. K-12 substr. MG1655':
            self.K = 42.00000
            self.BETA = 1.636217004872062
        else:
            self.K = 42.00000
            self.BETA = 1.636217004872062

    # Identify promoter with minimum dG_total (across many possible promoter states) for each TSS position in an inputted sequence.
    def predict(self, sequence, TSS_range):

        UPS_length = 24
        HEX35_length = 6
        UPS_HEX35_SPACER = 1
        SPACER_length_range = [15, 21]
        HEX10_length = 6
        DISC_length_range = [6, 11]
        ITR_length = 20

        MinPromoterSize = UPS_length + UPS_HEX35_SPACER + HEX35_length + SPACER_length_range[0] + HEX10_length + DISC_length_range[0] + ITR_length
        MaxPromoterSize = UPS_length + UPS_HEX35_SPACER + HEX35_length + SPACER_length_range[1] + HEX10_length + DISC_length_range[1] + ITR_length
        MinimumTSS = UPS_length + UPS_HEX35_SPACER + HEX35_length + SPACER_length_range[0] + HEX10_length + DISC_length_range[0]

        All_States = {}
        Min_States = {}

        #Specify fixed TSS range
        for TSS in range(TSS_range[0], TSS_range[1]):
            All_States[TSS] = {}
            for DISC_length in range(DISC_length_range[0],DISC_length_range[1]):

                if TSS - DISC_length >= 0 and TSS + ITR_length <= len(sequence):
                    tempdisc = sequence[ TSS - DISC_length : TSS  ]
                    tempITR  = sequence[ TSS : TSS + ITR_length]

                    for SPACER_length in range(SPACER_length_range[0], SPACER_length_range[1]):

                        if TSS - DISC_length - HEX10_length - SPACER_length - HEX35_length - UPS_length - UPS_HEX35_SPACER >= 0:
                            temp10     = sequence[ TSS - DISC_length - HEX10_length : TSS - DISC_length]
                            tempspacer = sequence[ TSS - DISC_length - HEX10_length - SPACER_length : TSS - DISC_length - HEX10_length ]
                            temp35     = sequence[ TSS - DISC_length - HEX10_length - SPACER_length - HEX35_length : TSS - DISC_length - HEX10_length - SPACER_length]
                            tempUP     = sequence[ TSS - DISC_length - HEX10_length - SPACER_length - HEX35_length - UPS_length - UPS_HEX35_SPACER:  TSS - DISC_length - HEX10_length - SPACER_length - HEX35_length - UPS_HEX35_SPACER]

                            dG_total, dG_apparent, dG_10, dG_35, dG_disc, dG_ITR, dG_ext10, dG_spacer, dG_UP = linear_free_energy_model(tempUP, temp35, tempspacer, temp10, tempdisc, tempITR, self.dg10_0, self.dg10_3, self.dg35_0, self.dg35_3, self.dmers, self.x10mers, self.spacers, self.model, self.inters)
                            dG_bind  = dG_10 + dG_35 + dG_spacer + dG_ext10 + dG_UP

                            Tx_rate = self.K * math.exp(- self.BETA * dG_total )

                            result = {'promoter_sequence' : sequence[TSS - DISC_length - HEX10_length - SPACER_length - HEX35_length - UPS_length - UPS_HEX35_SPACER : TSS + ITR_length ],
                                      'TSS' : TSS, 'UP' : tempUP, 'hex35' : temp35, 'spacer' : tempspacer, 'hex10' : temp10, 'disc' : tempdisc, 'ITR' : tempITR,
                                      'dG_total' : dG_total, 'dG_10' : dG_10, 'dG_35' : dG_35, 'dG_disc' : dG_disc, 'dG_ITR' : dG_ITR, 'dG_ext10' : dG_ext10, 'dG_spacer' : dG_spacer, 'dG_UP' : dG_UP, 'dG_bind' : dG_bind,
                                      'Tx_rate' : Tx_rate,
                                      'UP_position' : [TSS - DISC_length - HEX10_length - SPACER_length - HEX35_length - UPS_HEX35_SPACER - UPS_length,
                                                       TSS - DISC_length - HEX10_length - SPACER_length - HEX35_length - UPS_HEX35_SPACER],
                                      'hex35_position' : [TSS - DISC_length - HEX10_length - SPACER_length - HEX35_length, TSS - DISC_length - HEX10_length - SPACER_length],
                                      'spacer_position' : [TSS - DISC_length - HEX10_length - SPACER_length, TSS - DISC_length - HEX10_length],
                                      'hex10_position' : [TSS - DISC_length - HEX10_length, TSS - DISC_length],
                                      'disc_position' : [TSS - DISC_length, TSS]
                                      }

                            All_States[TSS][ (DISC_length, SPACER_length) ] = result
                            if TSS in Min_States:
                                if result['dG_total'] < Min_States[TSS]['dG_total']:  Min_States[TSS] = result
                            else:
                                Min_States[TSS] = result

        return (Min_States, All_States)

    def run(self, sequence, TSS_range = None):

        if TSS_range is None: TSS_range = [0, len(sequence)]

        self.sequence = sequence
        self.TSS_range = TSS_range
        self.TSS_range_rev = [len(sequence) - TSS_range[1], len(sequence) - TSS_range[0]]

        # print "self.TSS_range_rev: ", self.TSS_range_rev

        fwd_sequence = sequence
        rev_sequence = _revcomp(sequence)
        (Forward_Min_States, Forward_All_States) = self.predict(fwd_sequence, TSS_range = self.TSS_range)
        (Reverse_Min_States_Temp, Reverse_All_States_Temp) = self.predict(rev_sequence, TSS_range = self.TSS_range_rev)

        Reverse_Min_States = {}
        Reverse_All_States = {}
        # 0  <------>500 fwd 500 bp
        # 500<------>0   rev 500 bp
        #      275 TSS
        #      200-300
        #500-275 = 225
        #
        for TSS in Reverse_Min_States_Temp.keys():
            Reverse_Min_States[len(sequence) - TSS] = Reverse_Min_States_Temp[TSS]
            Reverse_All_States[len(sequence) - TSS] = Reverse_All_States_Temp[TSS]

        self.Forward_Predictions_per_TSS = Forward_Min_States
        self.Reverse_Predictions_per_TSS = Reverse_Min_States

    def output(self):
        output = {'organism' : self.organism,
                  'sigmaLevels' : self.sigmaLevels,
                  'K' : self.K,
                  'beta' : self.BETA,
                  'sequence' : self.sequence,
                  'TSS_range' : self.TSS_range,
                  'Forward_Predictions_per_TSS' : self.Forward_Predictions_per_TSS,
                  'Reverse_Predictions_per_TSS' : self.Reverse_Predictions_per_TSS
                }
        return output.copy()

    # Scan sequence left to right with no TSS information. Calc dG of all possible promoter configurations. Return promoter with minimum dG_total.
    def oldScan(self, inputSequence, preSeq, postSeq):
        seq_query = {}
        sequence = preSeq + inputSequence + postSeq

        #print "sequence: ", sequence

        # first 20 nt will be initial UP candidate
        for i in range(0,len(sequence)):
            tempUP = sequence[i:i+24]
            temp35 = sequence[i+25:25+i+6]  # leaves 1 nt between h35 and UPs

            # bounds defined by what was present during parameterization
            for j in range(15,21):
                tempspacer = sequence[i+25+6:25+i+6+j]
                temp10     = sequence[25+i+j+6:25+i+j+12]
                for k in range(6,11):
                    tempdisc  = sequence[25+i+j+12:25+i+j+12+k]
                    tempITR   =sequence[25+i+j+12+k:45+i+j+12+k]
                    if len(tempITR) < 20:
                        continue
                    else:
                        dG_total, dG_apparent, dg10, dg35, dg_disc, dg_ITR, dg_ext10, dg_spacer, dg_UP= linear_free_energy_model(tempUP, temp35, tempspacer, temp10, tempdisc, tempITR, self.dg10_0, self.dg10_3, self.dg35_0, self.dg35_3, self.dmers, self.x10mers, self.spacers, self.model, self.inters)
                        dG_bind  = dg10 + dg35 + dg_spacer + dg_ext10 + dg_UP
                        # dG_bind  = dg10 + dg_ext10 + dg_spacer + dg_UP
                        TSS_distance = i + len(tempUP) + len(temp35) + len(tempspacer) + len(temp10) + len(tempdisc)
                        # seq_query[(float(dG_bind), float(dG_total), TSS_distance)] = ((tempUP, temp35, tempspacer, temp10, tempdisc, tempITR),(dg10, dg35, dg_disc, dg_ITR, dg_ext10, dg_spacer, dg_UP))
                        seq_query[(float(dG_total), float(dG_apparent), TSS_distance)] = ((tempUP, temp35, tempspacer, temp10, tempdisc, tempITR),(dg10, dg35, dg_disc, dg_ITR, dg_ext10, dg_spacer, dg_UP))

        best = (collections.OrderedDict(sorted(seq_query.items())), min(seq_query.items(), key=operator.itemgetter(0)))
        return best, seq_query

if __name__ == "__main__":

    begin = time.time()
    sequence = input('Please provide sequence:\n').upper()
    calc = Promoter_Calculator()
    calc.run(sequence, TSS_range = [0, len(sequence)])
    output = calc.output()
    # Collect the output
    fwd_results = output['Forward_Predictions_per_TSS']
    rev_results = output['Reverse_Predictions_per_TSS']
    def _updator(dict, key, strand):
        dict[key]['strand'] = strand
    [_updator(fwd_results, key, "+") for key in fwd_results.keys()]
    [_updator(rev_results, key, "-") for key in rev_results.keys()]
    combined_results = {**fwd_results, **rev_results}
    # Sort the output by Tx_rate
    sorted_results = sorted(combined_results.items(), key=lambda x: x[1]['Tx_rate'], reverse=True)
    # Print the top 10 results
    for i in range(10):
        rprint(sorted_results[i])
    rprint("Elapsed Time: ", time.time() - begin, " seconds.")




"""
 Wrapper for Promoter_Calculator.py
"""

import time
import math
import importlib
from copy import copy
from .promoter_calculator import Promoter_Calculator, PromoCalcResults
from dataclasses import dataclass
from typing import Callable
from memory_profiler import profile
from .util import timer


if importlib.util.find_spec("progress") is not None:
    from progress.bar import Bar
else:
    Bar = None

def _progress_callback(tasks: int, description:str) -> Callable:
    # Update the progress bar
    if Bar is not None:
        bar = Bar(f'{description}: ', max=tasks)
    else:
        return None
    max = tasks
    itterations = 0

    def _callback():
        nonlocal itterations
        itterations += 1
        nonlocal max
        if itterations == max:
            bar.next()
            bar.finish()
            return None

        bar.next()
        return None
    return _callback

#@dataclass(frozen=True)
class PromoCalcResult(PromoCalcResults):
    pass


def promoter_calculator(sequence, threads=1, verbosity=1, callback=None, circular=False):
    begin = time.time()
    #sequence = "".join([random.choice(['A','G','C','T']) for x in range(100000)])
    sequence = sequence.upper()
    output = []


    # Helper function to get raw output
    def _run_calculator(target, start_pos, end_pos, fraction, callback=None):
        calc = Promoter_Calculator(threads=threads, verbosity=verbosity)
        calc.run(target, TSS_range = [0, len(target)], callback=callback)
        rev_results = calc.output()['Reverse_Predictions_per_TSS']
        fwd_results = calc.output()['Forward_Predictions_per_TSS']
        calculator_result = []
        for i in fwd_results.keys():
            fwd_results[i]['TSS_name'] = f"Fwd{fwd_results[i]['TSS']}"
            fwd_results[i]['TSS'] = start_pos + fwd_results[i]['TSS']
            fwd_results[i]['UP_position'][0] = start_pos + fwd_results[i]['UP_position'][0]
            fwd_results[i]['UP_position'][1] = start_pos + fwd_results[i]['UP_position'][1]
            fwd_results[i]['hex35_position'][0] = start_pos + fwd_results[i]['hex35_position'][0]
            fwd_results[i]['hex35_position'][1] = start_pos + fwd_results[i]['hex35_position'][1]
            fwd_results[i]['spacer_position'][0] = start_pos + fwd_results[i]['spacer_position'][0]
            fwd_results[i]['spacer_position'][1] = start_pos + fwd_results[i]['spacer_position'][1]
            fwd_results[i]['hex10_position'][0] = start_pos + fwd_results[i]['hex10_position'][0]
            fwd_results[i]['hex10_position'][1] = start_pos + fwd_results[i]['hex10_position'][1]
            fwd_results[i]['disc_position'][0] = start_pos + fwd_results[i]['disc_position'][0]
            fwd_results[i]['disc_position'][1] = start_pos + fwd_results[i]['disc_position'][1]
            fwd_results[i]['drop'] = False
            if type(fraction) != str:
                fwd_results[i]['fraction'] = fraction + 1
            fwd_results[i]['strand'] = '+'
            calculator_result.append(copy(fwd_results[i]))
        for i in rev_results.keys():
            rev_results[i]['TSS_name'] = f"Rev{rev_results[i]['TSS']}"
            rev_results[i]['TSS'] = end_pos - rev_results[i]['TSS']
            rev_results[i]['UP_position'][0] = end_pos - rev_results[i]['UP_position'][0]
            rev_results[i]['UP_position'][1] = end_pos - rev_results[i]['UP_position'][1]
            rev_results[i]['hex35_position'][0] = end_pos - rev_results[i]['hex35_position'][0]
            rev_results[i]['hex35_position'][1] = end_pos - rev_results[i]['hex35_position'][1]
            rev_results[i]['spacer_position'][0] = end_pos - rev_results[i]['spacer_position'][0]
            rev_results[i]['spacer_position'][1] = end_pos - rev_results[i]['spacer_position'][1]
            rev_results[i]['hex10_position'][0] = end_pos - rev_results[i]['hex10_position'][0]
            rev_results[i]['hex10_position'][1] = end_pos - rev_results[i]['hex10_position'][1]
            rev_results[i]['disc_position'][0] = end_pos - rev_results[i]['disc_position'][0]
            rev_results[i]['disc_position'][1] = end_pos - rev_results[i]['disc_position'][1]
            rev_results[i]['strand'] = '-'
            rev_results[i]['drop'] = False
            if type(fraction) != str:
                rev_results[i]['fraction'] = fraction + 1
            calculator_result.append(copy(rev_results[i]))
        for result in calculator_result:
            result['length'] = len(result['promoter_sequence'])

        # Filter out overlapping promoters

        calculator_result.sort(key=lambda x: x['TSS'])
        SEARCH_RANGE = 10  

        to_keep = []
        length = len(calculator_result)
        average = sum([x['dG_total'] for x in calculator_result]) / length

        # Legacy overlap filter
        for i in range(length):
            promotor = calculator_result.pop(0)
            if 'drop' not in promotor.keys():
                promotor['drop'] = False
            for _, other_promoter in enumerate(calculator_result):
                if other_promoter['TSS'] > promotor['TSS'] + SEARCH_RANGE:
                    continue
                elif promotor['fraction'] != other_promoter['fraction'] and promotor['fraction'] != 'junction' and other_promoter['fraction'] != 'junction':
                    continue
                elif promotor['strand'] != other_promoter['strand']:
                    continue
                for position_type in ['UP_position', 'hex35_position', 'spacer_position', 'hex10_position', 'disc_position']:
                    if promotor[position_type] == other_promoter[position_type]:
                        if promotor['dG_total'] < other_promoter['dG_total']:
                            promotor['drop'] = True
                            break
                        if promotor['dG_total'] > other_promoter['dG_total']:
                            other_promoter['drop'] = True
                            break
                        else:
                            promotor['drop'] = True
                            break
            if promotor['drop'] == False:
                to_keep.append(promotor)

        return to_keep, length, average
        


    # Run raw output of calculator, chunk if necessary
    fraction_length = 2500


    if len(sequence) >= 5*fraction_length and False:  # Disabled code
        if verbosity >= 2:
            print('Large input detected. Chunking sequence into smaller pieces.')
        fractions = math.floor(len(sequence)/fraction_length)
        remainder = len(sequence) % fraction_length
        if remainder > 0:
            fractions += 1
        else:
            # remainder onto the last run
            pass
        
        
        if not callback:
            if verbosity >= 1:
                callback = _progress_callback(description = "Running Promoter Calculator", tasks = (len(sequence) + 2*100*(fractions-1))*2)
            else:
                callback = lambda: None
        

        run_number = 0
        
        sample_hits = []
        sample_averages = []

        for i in range(fractions):
            if i == fractions-1:
                end_pos = len(sequence)
            else:
                end_pos = (i+1)*fraction_length+100
            if i == 0:
                start_pos = 0
            else:
                start_pos = i*fraction_length-100
            
            target_sequence = sequence[start_pos:end_pos]

            result, num_hits, average = _run_calculator(target_sequence, start_pos, end_pos, i, callback=callback)
            sample_hits.append(num_hits)
            sample_averages.append(average)
            output.extend(result)
            run_number += len(target_sequence)

        # Calculate weighted average
        num_hits = sum(sample_hits)
        average = sum([x*y for x,y in zip(sample_hits, sample_averages)]) / num_hits

    else:
        if not callback:
            if verbosity >= 1:
                callback = _progress_callback(description = "Running Promoter Calculator", tasks = len(sequence)*2)
            else:
                callback = lambda: None

        result, num_hits, average = _run_calculator(sequence, 0, len(sequence), 0, callback=callback)
        output.extend(result)

    # If the DNA is circular, examine the junction
    junction_results = {}
    if circular:
        if verbosity >= 2:
            print("Circular DNA detected. Examining junction.")
        # Get the junction
        junction = sequence[-100:] + sequence[:100]
        result, num_hits, average = _run_calculator(junction, len(sequence)-100, len(sequence)+100, "junction")
        # Filter out promotors that don't cross the junction
        for result in result:
            if result['strand'] == '+':
                bounds = [result['UP_position'][0], result['disc_position'][1]]
            if result['strand'] == '-':
                bounds = [result['disc_position'][0], result['UP_position'][1]]
            if bounds[0] <= 100 and bounds[1] >= 101:
                # Subtract 100 from the location to make the position appropriate relative to the rest of the sequence
                result['UP_position'][0] -= 100
                result['UP_position'][1] -= 100
                result['hex35_position'][0] -= 100
                result['hex35_position'][1] -= 100
                result['spacer_position'][0] -= 100
                result['spacer_position'][1] -= 100
                result['hex10_position'][0] -= 100
                result['hex10_position'][1] -= 100
                result['disc_position'][0] -= 100
                result['disc_position'][1] -= 100
                result['TSS'] -= 100
                if result['TSS'] < 0:
                    result['TSS'] += len(sequence)
                junction_results[f"{result['TSS']}{result['strand']}"] = result
    
        # Replace TSS with the actual position in the sequence
        for i in range(len(output)):
            position = output[i]['TSS']
            strand = output[i]['strand']
            if f"{position}{strand}" not in junction_results.keys():
                continue  # Ignore positions that are not hit by circularization
            # Take the more negative dG_total
            if output[i]['dG_total'] > junction_results[f"{position}{strand}"]['dG_total']:
                output[i] = junction_results[f"{position}{strand}"]

    # Find the best results
    output.sort(key=lambda x: x['Tx_rate'])
    output = output[::-1]
    return output



from setuptools import setup, find_packages

setup(
    name="promotercalculator",
    version="1.2",
    packages=["promoter_calculator"],
    package_dir={'promoter_calculator': 'promoter-calculator'},

    # Project uses reStructuredText, so ensure that the docutils get
    # installed or upgraded on the target machine
    install_requires=['Bio', 'numpy', 'scikit_learn', 'scipy'],

    # metadata to display on PyPI
    author="Cameron Roots",
    author_email="croots@utexas.edu",
    description="Salis Lab's promoter-calculator (Barrick Lab Fork)",
    keywords="DNA, RNA, promoter, promoter-calculator, calculator, Barrick, Salis, Biology, Bio",
    url="https://github.com/barricklab/promoter-calculator",   # project home page, if any
    project_urls={
        "Bug Tracker": "https://github.com/barricklab/promoter-calculator/issues",
        "Source Code": "https://github.com/barricklab/promoter-calculator",
    },
    include_package_data=True,
    classifiers=[
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)"
    ],
    entry_points={
        'console_scripts' : [
          'promoter-calculator = promoter_calculator.cli:main',
        ],
    }

    # could also include long_description, download_url, etc.
)



"""
Command line interface for the promoter-calculator.
"""

import argparse
import os
import csv
from copy import copy
from .wrapper import promoter_calculator
from dataclasses import asdict

def print_promo_calculator(output):
    print("#------------------------------------------------------#")
    print("Highest expressing promotors: \n")
    if not output:
        print("No promotors found.")
        return
    if len(output) > 25:
        num_results_to_print = 25
    else:
        num_results_to_print = len(output)
    for i in range(0, num_results_to_print):
        print(f"Highest Tx rate {i+1}: {output[i]['Tx_rate']} | Length: {output[i]['length']} | Position: {output[i]['UP_position']}{output[i]['strand']} | Fr {output[i]['fraction']}")


def main():
    """ CLI entry point """
    parser = argparse.ArgumentParser(description=f'Salis Lab Promoter Calculator')

    parser.add_argument(
        '-i', '--input',
        action='store',
        metavar='str/filepath',
        dest='i',
        required=False,
        type=str,
        help="Input DNA/RNA sequence.",
    )

    parser.add_argument(
        '-v', '--verbosity',
        action='store',
        metavar='int',
        dest='v',
        required=False,
        type=int,
        help="Verbosity level. 0 = silent, 1 = progress bar, 2 = detailed progress.",
        default=1
    )

    parser.add_argument(
        '-o', '--output',
        action='store',
        metavar='filepath',
        dest='o',
        required=False,
        type=str,
        help="Output file path.",
    )

    parser.add_argument(
        '-j', '--threads',
        action='store',
        metavar='int',
        dest='j',
        required=False,
        type=int,
        help="Number of threads to use. (Min 1)",
        default=1
    )

    parser.add_argument('-c',
        action='store_true',
        dest='c',
        required=False,
        help='Indecate the input is a circular sequence.',
        default=False)
    

    options = parser.parse_args()

    # Show help if no input is provided
    if not options.i:
        parser.print_help()
        return

    # Show help if threads is invalid
    if options.j < 1:
        parser.print_help()
        return

    if os.path.isfile(options.i):
        print("Input file is a file. Reading sequence from file.")
        with open(options.i, 'r') as f:
            sequence_lines = f.readlines()
            sequence = []
            for line in sequence_lines:
                line = line.strip()
                if ">" in line:
                    continue
                sequence.append(line)
            sequence = "".join(sequence)
            print(f"Sequence length: {len(sequence)}")
    else:
        sequence = options.i

    output = promoter_calculator(sequence, verbosity = options.v, threads = options.j, circular=options.c)

    if not options.o:
        print_promo_calculator(output)
    else:
        # Make a CSV with all the hits
        with open(options.o, 'w') as f:
            writer = csv.writer(f)
            writer.writerow(asdict(output[0]).keys())
            for row in output:
                writer.writerow(asdict(row).values())

if __name__ == "__main__":
    main()



from sklearn.preprocessing import OneHotEncoder, LabelEncoder
from Bio.Seq import Seq
from scipy import stats
import numpy as np
import itertools
import collections
import operator
from typing import Callable
import time
from functools import wraps

groove_access = {'CG' : 43,
                 'CA':  42, 'TG' : 42,  #revcomp
                 'GG':  42, 'CC' : 42,  #revcomp
                 'GC':  25,
                 'GA':  22, 'TC' : 22,  #revcomp
                 'TA':  14,
                 'AG':  9, 'CT' : 9,    #revcomp
                 'AA':  5, 'TT' : 5,    #revcomp
                 'AC':  4, 'GT' : 4,    #revcomp
                 'AT':  0
                }

stacking_dict = {'AA' :-2.41,
                 'AC':-2.06, 'CA' : -2.06, #reverse
                 'AG':-2.54, 'GA' : -2.54, #reverse
                 'AT':-2.25, 'TA' : -2.25, #reverse
                 'CC':-1.93,
                 'CG':-2.24, 'GC' : -2.24, #reverse
                 'CT':-2.01, 'TC' : -2.01, #reverse
                 'GG':-2.66,
                 'GT':-2.46, 'TG' : -2.46, #reverse
                 'TT':-1.82
                 }

average_twist = {'AA' : 35.6, 'TT' : 35.6, #revcomp
                 'AT':  32.1, #'AT' : 32.1, revcomp
                 'TA':  35.3, #'TA' : 35.3, revcomp
                 'GG':  33.65, 'CC' : 33.65, #revcomp
                 'GC':  40.2,
                 'CG':  29.9,
                 'AG':  27.9, 'CT' : 27.9, #revcomp
                 'GA':  36.8, 'TC' : 36.8, #revcomp
                 'AC':  34.0, 'GT' : 34.0, #revcomp
                 'CA':  34.4, 'TG' : 34.4 #revcomp
                 } # CR - Duplicate keys in dictionary?

# Sequence dependence of DNA bending rigidity, Geggier and Vologodskii
persistence   = {'AA' : 50.4, 'TT' : 50.4, #revcomp
                 'AC':  55.4, 'GT' : 55.4, #revcomp
                 'AG':  51.0, 'CT' : 51.0, #revcomp
                 'AT':  40.9, #'AT' : 40.9, revcomp
                 'CA':  46.7, 'TG' : 46.7, #revcomp
                 'CC':  41.7, 'GG' : 41.7, #revcomp
                 'CG':  56.0,
                 'GA':  54.4, 'TC' : 54.4, #revcomp
                 'GC':  44.6,
                 'TA':  44.7,
                } # CR - Duplicate keys in dictionary?
RNA_DNA_hybrids = {'TT': -1.0,
                   'TG': -2.1,
                   'TC': -1.8,
                   'TA': -0.9,
                   'GT': -0.9,
                   'GG': -2.1,
                   'GC': -1.7,
                   'GA': -0.9,
                   'CT': -1.3,
                   'CG': -2.7,
                   'CC': -2.9,
                   'CA': -1.1,
                   'AT': -0.6,
                   'AG': -1.5,
                   'AC': -1.6,
                   'AA': -0.2
                   }

DNA_DNA_hybrids = { 'AA':   -1.00, 'TT' : -1.00, #revcomp
                    'AT':   -0.88,
                    'TA':   -0.58,
                    'CA':   -1.45, 'TG' : -1.45, #revcomp
                    'GT':   -1.44, 'AC' : -1.44, #revcomp
                    'CT':   -1.28, 'AG' : -1.28, #revcomp
                    'GA':   -1.30, 'TC' : -1.30, #revcomp
                    'CG':   -2.17,
                    'GC':   -2.24,
                    'GG':   -1.42, 'CC' : -1.42  #revcomp
                  }


### This function calcs the hamming dist between 2 sequences of = len
def hamming(seq1, seq2):
    distance = 0
    L = len(seq1)
    if L == len(seq2):
        for i in range(L):
            if seq1[i] != seq2[i]:
                distance += 1
    else:
        print('Sequences are not of equal length')
        print(seq1, seq2)
        distance = None
    return distance

def get_rsquared(TXpred,TXempirical):
    slope, intercept, r_value, p_value, std_err = stats.mstats.linregress(TXpred,TXempirical)
    return r_value**2

### RETURN THE REVERSE COMPLIMENT OF A SEQ
revcomp_letters_util = {'U' : 'A', 'A' : 'T', 'G' : 'C', 'T' : 'A', 'C' : 'G'}
def reverse_complement(seq):
    return "".join([revcomp_letters_util[letter] for letter in seq[::-1] ])
    ## seq = Seq(seq)
    ## return seq.reverse_complement()

def AT_content(seq):
    ATcont = (seq.count('A') + seq.count('T'))/float(len(seq))*100
    return ATcont

def get_dg_matrices(parameters, encoder):
    dg_dict = {}
    mer_lis = []
    for entry in encoder.categories_[0]:
        mer_lis.append(entry)
    reference = mer_lis
    my_dict = dict(zip(reference,parameters))
    my_dict2 = collections.OrderedDict(sorted(my_dict.items()))
    sorted_by_val = sorted(my_dict2.items(), key=operator.itemgetter(1))
    for entry in sorted_by_val:
        dg_dict[entry[0]] = entry[1]
    return dg_dict

# Functions to calculate DNA or RNA properties
def calc_stacking_and_twist_energy(seq):
    #find stacking free enegy of spacer
    spacer_sfe  = 0
    spacer_twist = 0
    for j in range(0,len(seq)-1,2):
        mer = seq[j:j+2]
        spacer_sfe += stacking_dict[mer]
        spacer_twist += average_twist[mer]
        # try forward and reverse or compliment
        ## try:
        ##     spacer_sfe += stacking_dict[mer]
        ## except:
        ##     spacer_sfe += stacking_dict[mer[::-1]]
        ## try:
        ##     spacer_twist += average_twist[mer]
        ## except:
        ##    spacer_twist += average_twist[reverse_complement(mer)]

    return spacer_sfe, spacer_twist

# local DNA element rigidity
def calc_rigidity(seq):
    rigidity = 0
    for m in range(0, len(seq), 2):
        mer = seq[m:m+2]
        rigidity += persistence[mer]
        ## try:
        ##     rigidity += persistence[mer]
        ## except:
        ##    rigidity += persistence[reverse_complement(mer)]

    rigidity = rigidity / len(seq)
    return rigidity

def calc_disc_melting(seq, downstream):
    dg_dna = 0
    for j in range(0,len(seq[0:len(seq)])-1,2):
        mer = seq[j:j+2]
        if len(mer) < 2:    mer = mer + downstream[0]
        if mer in DNA_DNA_hybrids:  dg_dna += DNA_DNA_hybrids[mer]
        dg_dna += DNA_DNA_hybrids[mer]
        ## try:
        ##     dg_dna += DNA_DNA_hybrids[mer]
        ## except:
        ##    dg_dna += DNA_DNA_hybrids[reverse_complement(mer)]

    return dg_dna

# find free enegy of DNA:DNA and RNA:DNA hybrids in ITR
def calc_DNA_RNA_hybrid_energy(seq):
    dg_dna = 0
    dg_rna = 0
    for j in range(0,len(seq[0:15])-1,2): # 15 is the length of long abortive transcripts
        mer = seq[j:j+2]
        dg_dna += DNA_DNA_hybrids[mer]
        # try forward and reverse or compliment
        ## try:
        ##     dg_dna += DNA_DNA_hybrids[mer]
        ## except:
        ##    dg_dna += DNA_DNA_hybrids[reverse_complement(mer)]
        dg_rna += RNA_DNA_hybrids[mer]

    dg_hybrid    = dg_dna - dg_rna
    return dg_dna,dg_rna,dg_hybrid

def calc_groove_width(seq):
    groove_width = 0
    for l in range(0, len(seq), 2):
        mer1 = seq[l:l+2]
        groove_width += groove_access[mer1]
        ## try:
        ##    groove_width+= groove_access[mer1]
        ## except:
        ##    groove_width+= groove_access[reverse_complement(mer1)]
    return groove_width

def length_encoders(lower_bound, upper_bound):
    length_array =[]
    for x in range(lower_bound, upper_bound+1):
        length_array.append(str(x))
    length_array = np.array(length_array)

    #one hot encode the array of length into bit vectors
    onehot_encoder = OneHotEncoder(sparse_output=False)
    onehot_encoder.fit(length_array.reshape(-1, 1))
    return onehot_encoder

def kmer_encoders(k):
    # k is the kmer size for DNA sequence
    # this will give bit vectors for 2mers or 3mers for us
    kmer_array =[]
    nts = ['A','T','G','C']
    for output in itertools.product(nts, repeat = k):
        kmer_array.append(''.join(output))
    kmer_array = np.array(kmer_array)

    #one hot encode the sequence
    onehot_encoder = OneHotEncoder(sparse_output=False)
    onehot_encoder.fit(kmer_array.reshape(-1, 1))
    return onehot_encoder

def timer(function: Callable) -> Callable:
    """Print the runtime of the decorated function"""
    @wraps(function)
    def wrapper_timer(*args, **kwargs):
        start_time = time.perf_counter()  # 1
        value = function(*args, **kwargs)
        end_time = time.perf_counter()  # 2
        run_time = end_time - start_time  # 3
        print(f"Finished {function.__name__!r} in {run_time:.4f} secs")
        return value
    return wrapper_timer


#EOF
